import re
from typing import Dict, List, Optional, Tuple, Union
from warnings import warn

import attrs
import neuroglancer
from numpy import integer, issubdtype

from . import utils
from .base import SEGMENTATION_LAYER_TYPES, EasyViewerBase

DEFAULT_SKELETON_SHADER = """void main() {
  emitDefault();
}
"""


def TagToolFactory(number_tags: int):
    for ii in range(number_tags):

        @neuroglancer.viewer_state.export_tool
        class TagTool(neuroglancer.viewer_state.Tool):
            __slots__ = ()
            TOOL_TYPE = f"tagTool_tag{ii}"


TagToolFactory(24)


@attrs.define
class AnnotationTag:
    id: str = None
    type: str = attrs.field(default="uint8")
    tag: str = None


def _make_annotation_properties(annotations, tag_base_number=0):
    "Take a list of annotations and build up the dictionary needed"
    properties = []
    for ii, an in enumerate(annotations):
        properties.append(
            attrs.asdict(AnnotationTag(id=f"tag{ii+tag_base_number}", tag=an))
        )
    return properties


def _make_bindings(properties, bindings=None):
    "Make a dictinary describing key bindings for tags"
    if bindings is None:
        bindings = ["Q", "W", "E", "R", "T", "A", "S", "D", "F", "G"]
    if len(properties) > len(bindings):
        raise ValueError("Too many properties for bindings")
    tool_bindings = {}
    for bind, prop in zip(bindings, properties):
        tool_bindings[bind] = f"tagTool_{prop['id']}"
    return tool_bindings


def nanometer_dimension(resolution):
    return neuroglancer.CoordinateSpace(
        names=["x", "y", "z"],
        scales=resolution,
        units="nm",
    )


class UnservedViewer(neuroglancer.viewer_base.UnsynchronizedViewerBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._default_viewer_url = kwargs.pop(
            "default_viewer_url", utils.default_mainline_neuroglancer_base
        )

    def get_server_url(self):
        return self._default_viewer_url


class EasyViewerMainline(UnservedViewer, EasyViewerBase):
    def __init__(self, **kwargs):
        super(UnservedViewer, self).__init__(**kwargs)
        super(EasyViewerBase, self).__init__(**kwargs)

    def load_url(self, url) -> None:
        "Parse a neuroglancer state based on URL and load it into the state"
        state = neuroglancer.parse_url(url)
        self.set_state(state)

    def _ImageLayer(self, source, **kwargs):
        if isinstance(source, str):
            source = utils.parse_graphene_header(source, target="mainline")
        return neuroglancer.viewer_state.ImageLayer(source=source, **kwargs)

    def _SegmentationLayer(self, source, **kwargs):
        if isinstance(source, str):
            source = utils.parse_graphene_header(source, target="mainline")
        elif isinstance(source, list):
            source = [utils.parse_graphene_header(s, target="mainline") for s in source]
        return neuroglancer.viewer_state.SegmentationLayer(source=source, **kwargs)

    def _AnnotationLayer(self, *args, **kwargs):
        return neuroglancer.viewer_state.LocalAnnotationLayer(*args, **kwargs)

    def set_resolution(self, resolution) -> None:
        with self.txn() as s:
            s.dimensions = nanometer_dimension(resolution)

    def set_state_server(self, state_server) -> None:
        Warning("State server is set by neuroglancer deployment for this viewer type.")

    def add_annotation_layer(
        self,
        layer_name=None,
        color=None,
        linked_segmentation_layer=None,
        filter_by_segmentation=False,
        brackets_show_segmentation=True,
        selection_shows_segmentation=True,
        tags=None,
        key_bindings=None,
    ) -> None:
        if layer_name is None:
            layer_name = "annos"
        if layer_name in [l.name for l in self.state.layers]:
            raise ValueError("Layer name already exists")

        if filter_by_segmentation:
            filter_by_segmentation = ["segments"]
        else:
            filter_by_segmentation = []

        if linked_segmentation_layer is not None:
            linked_segmentation_layer = {"segments": linked_segmentation_layer}
        else:
            linked_segmentation_layer = {}

        with self.txn() as s:
            new_layer = self._AnnotationLayer(
                dimensions=self.state.dimensions,
                linked_segmentation_layer=linked_segmentation_layer,
                filter_by_segmentation=filter_by_segmentation,
            )
            s.layers.append(name=layer_name, layer=new_layer)
            if color is not None:
                s.layers[layer_name].annotationColor = utils.parse_color(color)

        if tags:
            self.add_annotation_tags(
                layer_name=layer_name, tags=tags, key_bindings=key_bindings
            )

    def _get_number_of_tags(self, layer_name):
        anno_props = self.state.layers[layer_name].annotationProperties
        num_tags = 0
        for prop in anno_props:
            if re.match(r"^tag\d+$", prop.get("id", "")):
                num_tags += 1
        return num_tags

    def add_annotation_tags(self, layer_name, tags, key_bindings=None):
        tag_base_number = self._get_number_of_tags(layer_name)
        props = _make_annotation_properties(tags, tag_base_number)
        bindings = _make_bindings(props, key_bindings)
        with self.txn() as s:
            s.layers[layer_name].annotationProperties = props
            s.layers[layer_name].tool_bindings = bindings

    def as_url(
        self,
        prefix: Optional[str] = None,
        as_html: Optional[bool] = False,
        link_text: Optional[str] = "Neuroglancer Link",
    ) -> str:
        if prefix is None:
            prefix = utils.default_mainline_neuroglancer_base
        ngl_url = neuroglancer.to_url(self.state, prefix=prefix)
        if as_html:
            return f'<a href="{ngl_url}" target="_blank">{link_text}</a>'
        else:
            return ngl_url

    def set_selected_layer(self, layer_name):
        with self.txn() as s:
            s.selected_layer = neuroglancer.SelectedLayerState(
                {"visible": True, "layer": layer_name}
            )

    def add_contrast_shader(self, layer_name, black=0, white=255):
        if isinstance(black, float) and black < 1:
            black = int(black * 255)
        if isinstance(white, float) and white < 1:
            white = int(white * 255)
        with self.txn() as s:
            s.layers[
                layer_name
            ].shader_controls = neuroglancer.viewer_state.ShaderControls(
                {"normalized": {"range": [black, white]}}
            )

    def select_annotation(self, layer_name, annotation_id):
        # self.set_selected_layer(layer_name)
        raise Warning("Annotation selection is not supported by this viewer type.")

    def assign_colors(self, layer_name, seg_colors):
        with self.txn() as s:
            s.layers[layer_name].segmentColors = seg_colors

    def add_selected_objects(
        self,
        segmentation_layer: str,
        oids: List[int],
        colors: Optional[Union[List, Dict]] = None,
        invisible_oids: Optional[List[int]] = None,
    ) -> None:
        if issubdtype(type(oids), integer):
            oids = [oids]
        if invisible_oids is None:
            invisible_oids = []
        with self.txn() as s:
            for oid in oids:
                s.layers[segmentation_layer].segments.add(str(oid))
            for oid in invisible_oids:
                s.layers[segmentation_layer].segments.add(f"!{oid}")
        if colors is not None:
            if isinstance(colors, dict):
                self.assign_colors(segmentation_layer, colors)
            elif len(colors) == len(oids):
                seg_colors = {oid: clr for oid, clr in zip(oids, colors)}
                self.assign_colors(segmentation_layer, seg_colors)

    def add_segmentation_layer(self, layer_name, source, **kwargs):
        """Add segmentation layer to viewer instance.

        Attributes:
            layer_name (str): name of layer to be displayed in neuroglancer ui.
            source (str): source of neuroglancer segment layer
        """
        with self.txn() as s:
            s.layers[layer_name] = self._SegmentationLayer(source=source, **kwargs)

    def append_source_to_segmentation_layer(self, layer_name, source):
        """Append source or sources to an existing segmentation layer.

        Parameters
        ----------
        layer_name : str
            Name of an existing layer
        source : str or list of str
            Source or sources to add to the layer. Can be segment properties or skeleton sources, for example.
        """
        if isinstance(source, str):
            source = [source]
        source = [utils.parse_graphene_header(s, target="mainline") for s in source]

        with self.txn() as s:
            for src in source:
                s.layers[layer_name].source.append({"url": src})

    def add_skeleton_source(self, layer_name, source, shader_text=None):
        if shader_text is None:
            shader_text = DEFAULT_SKELETON_SHADER
        with self.txn() as s:
            s.layers[layer_name].source.append({"url": source})
            s.layers[layer_name].skeletonShader = shader_text

    def set_skeleton_shader(self, layer_name, shader_text=None):
        if shader_text is None:
            shader_text = DEFAULT_SKELETON_SHADER
        with self.txn() as s:
            s.layers[layer_name].skeletonShader = shader_text

    def set_view_options(
        self,
        show_slices: Optional[bool] = None,
        layout: Optional[str] = None,
        show_axis_lines: Optional[bool] = None,
        show_scale_bar: Optional[bool] = None,
        orthographic: Optional[bool] = None,
        position: Optional[Tuple[float]] = None,
        zoom_image: Optional[float] = None,
        zoom_3d: Optional[float] = None,
        background_color: Optional[Tuple[float]] = None,
    ) -> None:
        with self.txn() as s:
            if show_slices is not None:
                s.showSlices = show_slices
            if layout is not None:
                s.layout = layout
            if show_axis_lines is not None:
                s.showAxisLines = show_axis_lines
            if show_scale_bar is not None:
                s.showScaleBar = show_scale_bar
            if orthographic is not None:
                s.layout.orthographic_projection = orthographic
            if position is not None:
                s.position = position
            if zoom_image is not None:
                s.crossSectionScale = zoom_image
            if zoom_3d is not None:
                s.projectionScale = zoom_3d
            if background_color is not None:
                s.projectionBackgroundColor = utils.parse_color(background_color)

    def set_segmentation_view_options(
        self,
        layer_name: str,
        alpha_selected: Optional[float] = None,
        alpha_3d: Optional[float] = None,
        alpha_unselected: Optional[float] = None,
        silhouette_value: Optional[float] = None,
        **kwargs,
    ):
        if self.state.layers[layer_name].type not in SEGMENTATION_LAYER_TYPES:
            raise ValueError("Layer is not a segmentation layer")
        with self.txn() as s:
            l = s.layers[layer_name]
            if alpha_selected is not None:
                l.selectedAlpha = alpha_selected
            if alpha_3d is not None:
                l.objectAlpha = alpha_3d
            if alpha_unselected is not None:
                l.notSelectedAlpha = alpha_unselected
            if silhouette_value is not None:
                l.meshSilhouetteRendering = silhouette_value

    def set_timestamp(
        self,
        layer_name,
        timestamp: Optional[int] = None,
    ):
        if timestamp is not None:
            warn("Timestamp setting is not yet enabled for this viewer type.")

    def set_multicut_points(
        self,
        layer_name,
        seg_id,
        points_red,
        points_blue,
        supervoxels_red=None,
        supervoxels_blue=None,
        focus=True,
    ):
        Warning("Setting multicut is not yet enabled for this viewer type")

    @staticmethod
    def point_annotation(
        point,
        id=None,
        description=None,
        linked_segmentation=None,
        annotation_properties=None,
        **kwargs,
    ):
        segments = utils.omit_nones(linked_segmentation)
        if id is None:
            id = neuroglancer.random_token.make_random_token()
        return neuroglancer.PointAnnotation(
            point=list(point),
            id=id,
            description=description,
            segments=[[int(x) for x in segments]],
            props=annotation_properties,
        )

    @staticmethod
    def line_annotation(
        pointA,
        pointB,
        id=None,
        description=None,
        linked_segmentation=None,
        annotation_properties=None,
        **kwargs,
    ):
        segments = utils.omit_nones(linked_segmentation)
        if id is None:
            id = neuroglancer.random_token.make_random_token()
        return neuroglancer.LineAnnotation(
            point_a=list(pointA),
            point_b=list(pointB),
            id=id,
            description=description,
            segments=[[int(x) for x in segments]],
            props=annotation_properties,
        )

    @staticmethod
    def ellipsoid_annotation(
        center,
        radii,
        id=None,
        description=None,
        linked_segmentation=None,
        annotation_properties=None,
        **kwargs,
    ):
        segments = utils.omit_nones(linked_segmentation)
        if id is None:
            id = neuroglancer.random_token.make_random_token()
        return neuroglancer.EllipsoidAnnotation(
            center=list(center),
            radii=list(radii),
            id=id,
            description=description,
            segments=[[int(x) for x in segments]],
            props=annotation_properties,
        )

    @staticmethod
    def bounding_box_annotation(
        pointA,
        pointB,
        id=None,
        description=None,
        linked_segmentation=None,
        annotation_properties=None,
        **kwargs,
    ):
        segments = utils.omit_nones(linked_segmentation)
        if id is None:
            id = neuroglancer.random_token.make_random_token()
        return neuroglancer.AxisAlignedBoundingBoxAnnotation(
            point_a=list(pointA),
            point_b=list(pointB),
            id=id,
            description=description,
            segments=[[int(x) for x in segments]],
            props=annotation_properties,
        )

    @staticmethod
    def group_annotations(
        annotations,
        source=None,
        id=None,
        return_all=True,
        gather_linked_segmentations=True,
        share_linked_segmentations=False,
        children_visible=True,
        **kwargs,
    ):
        Warning("Annotation groups are not yet supported by this viewer type.")



from __future__ import annotations

from typing import Literal, Optional

from IPython.display import HTML

from nglui.easyviewer import EasyViewer
from nglui.easyviewer.ev_base.nglite.json_utils import encode_json

from ..site_utils import get_default_config, neuroglancer_url

DEFAULT_VIEW_KWS = {
    "layout": "xy-3d",
    "zoom_image": 1,
    "show_slices": False,
    "zoom_3d": 5000,
}


class StateBuilder:
    """A class for schematic mapping data frames into neuroglancer states.

    Parameters
    ----------
    layers : list of nglui.statebuilder.layers.LayerConfigBase, optional
        List of layers to add. Defaults to [].
    base_state : dict, optional
        JSON state to add to. Defaults to None.
    url_prefix : str, optional
        http(s) path to Neuroglancer deployment to use. Defaults to None,
        which will use https://neuromancer-seung-import.appspot.com
    state_server : str, optional
        State server to post links to. Defaults to None.
    resolution : list of float, optional
        3-element vector controlling the viewer resolution. Defaults to None.
        If None and a client is set, uses the client viewer resolution.
    view_kws : dict, optional
        Dictionary controlling view parameters. Defaults to {}.

        Keys are:
            show_slices : bool, optional
                Sets if slices are shown in the 3d view. Defaults to False.
            layout : str, optional
                `xy-3d`/`xz-3d`/`yz-3d` (sections plus 3d pane),
                `xy`/`yz`/`xz`/`3d` (only one pane), or `4panel` (all panes).
                Default is `xy-3d`.
            show_axis_lines : bool, optional
                Determines if the axis lines are shown in the middle of each view.
            show_scale_bar : bool, optional
                Toggles showing the scale bar.
            orthographic : bool, optional
                Toggles orthographic view in the 3d pane.
            position : list of float, optional (length-3)
                Determines the centered location.
            zoom_image : float, optional
                Zoom level for the imagery in units of nm per voxel. Defaults to 8.
            zoom_3d : float, optional
                Zoom level for the 3d pane. Defaults to 2000. Smaller numbers are
                more zoomed in.
            background_color : str or list of float, optional
                Sets the background color of the 3d view. Arguments can be rgb
                values, hex colors, or named web colors. Defaults to black.
    client : caveclient.CAVEclient, optional
        A caveclient to get defaults from. Defaults to None.
    target_site : str, optional
        Target Neuroglancer category: either "seunglab" or "mainline"/"cave-explorer"/"spelunker". Defaults to None.
        Will be looked up automatically based on ngl_url, if used.
    """

    def __init__(
        self,
        layers: list = [],
        base_state: Optional[dict] = None,
        url_prefix: Optional[str] = None,
        state_server: Optional[str] = None,
        resolution: Optional[list] = None,
        view_kws: dict = {},
        client: Optional[caveclient.CAVEclient] = None,
        target_site: Optional[
            Literal["seunglab", "mainline", "cave-explorer", "spelunker"]
        ] = None,
        config_key: Optional[str] = None,
    ):
        _config = get_default_config(config_key)
        if client is None:
            client = _config["caveclient"]
        if target_site is None:
            target_site = _config["target_site"]
        if client is not None:
            if state_server is None:
                state_server = client.state.state_service_endpoint
            if resolution is None:
                resolution = client.info.viewer_resolution().tolist()
        self._client = client
        self._state_server = state_server
        self._base_state = base_state
        self._layers = layers
        self._resolution = resolution
        self._target_site = target_site
        self._url_prefix = url_prefix

        base_kws = DEFAULT_VIEW_KWS.copy()
        base_kws.update(view_kws)
        self._view_kws = base_kws

    @property
    def viewer(self):
        return self._temp_viewer

    def _reset_state(self, base_state=None, target_site=None):
        """
        Resets the neuroglancer state status to a default viewer.
        """
        if base_state is None:
            base_state = self._base_state
        if target_site is None:
            target_site = self._target_site
        self._temp_viewer = EasyViewer(target_site=self._target_site)
        self._temp_viewer.set_state(base_state)

    def initialize_state(self, base_state=None, target_site=None):
        """Generate a new Neuroglancer state with layers as needed for the schema.

        Parameters
        ----------
        base_state : str, optional
            Optional initial state to build on, described by its JSON. By default None.
        """
        self._reset_state(base_state, target_site=target_site)

        if self._state_server is not None:
            self._temp_viewer.set_state_server(self._state_server)

        if self._resolution is not None:
            self._temp_viewer.set_resolution(self._resolution)

        self._temp_viewer.set_view_options(**self._view_kws)
        for l in self._layers:
            l._add_layer(self._temp_viewer)

    def handle_positions(self, data):
        for l in self._layers[::-1]:
            pos = l._set_view_options(
                self._temp_viewer, data, viewer_resolution=self._resolution
            )
            if pos is not None:
                break

    def render_state(
        self,
        data: Optional[pandas.DataFrame] = None,
        base_state: Optional[dict] = None,
        return_as: Literal["url", "viewer", "html", "json", "dict", "short"] = "url",
        url_prefix: Optional[str] = None,
        link_text: str = "Neuroglancer Link",
        target_site: Optional[str] = None,
        client: Optional[caveclient.CAVEclient] = None,
        config_key: Optional[str] = None,
    ):
        """Build a Neuroglancer state out of a DataFrame.

        Parameters
        ----------
        data : pandas.DataFrame, optional
            DataFrame to use as a point source. By default None, for which
            it will return only the base_state and any fixed values.
        base_state : dict, optional
            Initial state to build on, expressed as Neuroglancer JSON. By default None
        return_as : ['url', 'viewer', 'html', 'json', 'dict', 'short'], optional
            Choice of output types. Note that if a viewer is returned, the state is not reset.
                url : Returns the raw url describing the state
                viewer : Returns an EasyViewer object holding the state information
                html : Returns an HTML link to the url, useful for notebooks.
                json : Returns a JSON string describing the state.
                dict : Returns a dict version of the JSON state.
                short : Posts the state to the state server and returns a short link. Requires a client.
            By default 'url'
        url_prefix : str, optional
            Neuroglancer URL prefix to use. By default None, for which it will open with the
            class default.
        link_text : str, optional
            Text to use for the link when returning as html, by default 'Neuroglancer Link'
        target_site : str, optional
            Target Neuroglancer category: either "seunglab" or one of "mainline"/"cave-explorer"/"spelunker". Defaults to None.
            Will be looked up automatically based on ngl_url, if used and a client is set.
        client : caveclient.CAVEclient, optional
            A caveclient to get defaults from. Defaults to None, which falls back on the statebuilder.
        config_key : str, optional
            Name of the configuration to use, by default "default"

        Returns
        -------
        string or neuroglancer.Viewer
            A link to or viewer for a Neuroglancer state with layers, annotations, and selected objects determined by the data.
        """
        if config_key is not None:
            _config = get_default_config(config_key)
            if client is None:
                client = _config["caveclient"]
            if target_site is None:
                target_site = _config["target_site"]
        else:
            if client is None:
                client = self._client
            if target_site is None:
                target_site = self._target_site
        if base_state is None:
            base_state = self._base_state
        if url_prefix is None:
            url_prefix = self._url_prefix

        url_prefix = neuroglancer_url(url_prefix, target_site, config_key=config_key)
        self.initialize_state(base_state=base_state, target_site=target_site)
        self.handle_positions(data)

        self._render_layers(
            data,
            client=client,
        )

        self._temp_viewer.set_view_options(**self._view_kws)

        if return_as == "viewer":
            return self.viewer
        elif return_as == "url":
            url = self._temp_viewer.as_url(prefix=url_prefix)
            return url
        elif return_as == "html":
            out = self._temp_viewer.as_url(
                prefix=url_prefix, as_html=True, link_text=link_text
            )
            out = HTML(out)
            return out
        elif return_as == "dict":
            out = self._temp_viewer.state.to_json()
            return out
        elif return_as == "json":
            out = self._temp_viewer.state.to_json()
            return encode_json(out)
        elif return_as == "short":
            if client is None:
                raise ValueError("Cannot generate short link without a client")
            state_id = client.state.upload_state_json(self._temp_viewer.state.to_json())
            short_link = client.state.build_neuroglancer_url(
                state_id,
                ngl_url=url_prefix,
                target_site=target_site.replace("spelunker", "mainline"),
            )
            return short_link
        else:
            raise ValueError("No appropriate return type selected")

    def _render_layers(
        self,
        data,
        client=None,
    ):
        if client is None:
            client = self._client
        # Inactivate all layers except last.
        found_active = False
        for l in self._layers[::-1]:
            if l.active and not found_active:
                found_active = True
            else:
                l.active = False
        anno_dict = {}
        for layer in self._layers:
            anno_dict[layer.name] = layer._render_layer(
                self._temp_viewer,
                data,
                viewer_resolution=self._resolution,
                return_annos=True,
                client=client,
            )
        self._temp_viewer.add_multilayer_annotations(anno_dict)


class ChainedStateBuilder:
    """Builds a collection of states that sequentially add annotations based on a sequence of dataframes.

    Parameters
    ----------
    statebuilders : list
        List of DataStateBuilders, in same order as dataframes will be passed
    """

    def __init__(self, statebuilders):
        self._statebuilders = statebuilders
        if len(self._statebuilders) == 0:
            raise ValueError("Must have at least one statebuilder")

    def render_state(
        self,
        data_list=None,
        base_state=None,
        return_as="url",
        url_prefix=None,
        link_text="Neuroglancer Link",
        target_site=None,
        client=None,
        config_key: Optional[str] = None,
    ):
        """Generate a single neuroglancer state by addatively applying an ordered collection of
        dataframes to an collection of StateBuilder renders.
        Parameters
        ----------

        data_list : Collection of DataFrame. The order must match the order of StateBuilders
                    contained in the class on creation.
        base_state : JSON neuroglancer state (optional, default is None).
                        Used as a base state for adding everything else to.
        return_as: ['url', 'viewer', 'html', 'json', 'short']. optional, default='url'.
                    Sets how the state is returned. Note that if a viewer is returned,
                    the state is not reset to default.
        url_prefix: string, optional (default is https://neuromancer-seung-import.appspot.com).
                    Overrides the default neuroglancer url for url generation.
        link_text: string, optional (default is 'Neuroglancer Link').
                    Text to use for the link when returning as html.
        target_site: string, optional (default is None).
                        Target Neuroglancer category: either "seunglab" or "mainline"/"cave-explorer"/"spelunker".
                        Will be looked up automatically based on url_prefix, if used.
        client: caveclient.CAVEclient, optional (default is None).
                    A caveclient to get defaults from. Defaults to None, which falls back on the last statebuilder.
        config_key: string, optional (default is None).
                    Name of the configuration to use.
        """
        if client is None:
            client = self._statebuilders[-1]._client
        if target_site is None:
            target_site = self._statebuilders[-1]._target_site
        if url_prefix is None:
            url_prefix = self._statebuilders[-1]._url_prefix

        url_prefix = neuroglancer_url(url_prefix, target_site, config_key=config_key)

        if data_list is None:
            data_list = len(self._statebuilders) * [None]

        if len(data_list) != len(self._statebuilders):
            raise ValueError("Must have as many dataframes as statebuilders")

        temp_state = base_state
        for builder, data in zip(self._statebuilders[:-1], data_list[:-1]):
            temp_state = builder.render_state(
                data=data,
                base_state=temp_state,
                return_as="dict",
                url_prefix=url_prefix,
                target_site=target_site,
            )
        last_builder = self._statebuilders[-1]
        return last_builder.render_state(
            data=data_list[-1],
            base_state=temp_state,
            return_as=return_as,
            url_prefix=url_prefix,
            link_text=link_text,
            target_site=target_site,
            client=client,
        )



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import contextlib
import copy
import threading

from .random_token import make_random_token


class ConcurrentModificationError(RuntimeError):
    pass


class ChangeNotifier:
    def __init__(self):
        self.__changed_callbacks = set()
        self.change_count = 0
        self.__lock = threading.Lock()

    def add_changed_callback(self, callback):
        """Registers a callback to be invoked when the state changes.

        The callback is invoked immediately with no arguments.  The callback must not block.
        """
        with self.__lock:
            self.__changed_callbacks.add(callback)

    def remove_changed_callback(self, callback):
        with self.__lock:
            self.__changed_callbacks.remove(callback)

    def _dispatch_changed_callbacks(self):
        with self.__lock:
            self.change_count += 1
            for callback in self.__changed_callbacks:
                callback()


class TrackableState(ChangeNotifier):
    def __init__(self, wrapper_type, transform_state=None):
        super(TrackableState, self).__init__()
        self._raw_state = {}
        self._lock = threading.RLock()
        self._generation = make_random_token()
        self._wrapped_state = None
        self._wrapper_type = wrapper_type
        if transform_state is None:

            def transform_state_function(new_state):
                if isinstance(new_state, wrapper_type):
                    return new_state.to_json()
                return new_state

            transform_state = transform_state_function
        self._transform_state = transform_state

    def set_state(self, new_state, generation=None, existing_generation=None):
        with self._lock:
            if (
                existing_generation is not None
                and self._generation != existing_generation
            ):
                raise ConcurrentModificationError
            new_state = self._transform_state(new_state)
            if new_state != self._raw_state or (
                generation is not None and generation != self._generation
            ):
                if generation is None:
                    generation = make_random_token()
                self._raw_state = new_state
                self._wrapped_state = None
                self._generation = generation
                self._dispatch_changed_callbacks()
            return self._generation

    @property
    def state_and_generation(self):
        with self._lock:
            return (self.state, self.state_generation)

    @property
    def raw_state_and_generation(self):
        with self._lock:
            return (self.raw_state, self.state_generation)

    @property
    def state_generation(self):
        return self._generation

    @property
    def raw_state(self):
        return self._raw_state

    @property
    def state(self):
        with self._lock:
            wrapped_state = self._wrapped_state
            if wrapped_state is None:
                wrapped_state = self._wrapped_state = self._wrapper_type(
                    self._raw_state, _readonly=True
                )
            return wrapped_state

    @contextlib.contextmanager
    def txn(self, overwrite=False, lock=True):
        """Context manager for a state modification transaction."""
        if lock:
            self._lock.acquire()
        try:
            new_state, existing_generation = self.state_and_generation
            new_state = copy.deepcopy(new_state)
            yield new_state
            if overwrite:
                existing_generation = None
            self.set_state(new_state, existing_generation=existing_generation)
        finally:
            if lock:
                self._lock.release()

    def retry_txn(self, func, retries=10):
        for retry in range(retries):
            try:
                with self.txn(lock=False) as s:
                    return func(s)
            except ConcurrentModificationError:
                if retry + 1 < retries:
                    pass
                raise

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.state)



from .helpers import from_client, make_neuron_neuroglancer_link
from .layers import AnnotationLayerConfig, ImageLayerConfig, SegmentationLayerConfig
from .mappers import (
    BoundingBoxMapper,
    LineMapper,
    PointMapper,
    SelectionMapper,
    SphereMapper,
    SplitPointMapper,
)
from .statebuilder import ChainedStateBuilder, StateBuilder
